#include "image.h"


Image::Image(int w, int h){
  width=w;
  height=h;
  image=(int*) _TIFFmalloc ( w  * h * sizeof(int));
  //std::cout << width << "+" << height << std::endl;
}
 

  	
Image::Image(char* fichier) {
  TIFF* sourceFile=TIFFOpen(fichier,"r");
  int*   memoryS         = (int*)NULL;
  int*   memoryD         = (int*)NULL;
  if (sourceFile == (TIFF*)NULL) {
    printf ("Could not open file named %s for reading.\n", fichier);
  }
  int res = LoadPixels (sourceFile, &memoryS, &width, &height);
  
  if (res == 0) {   
    SwapBuffer (&memoryD, memoryS, width, height); 
  }
  image=memoryD;
  TIFFClose (sourceFile);
  //std::cout << width << "+" << height << std::endl;
  
}

int Image::getWidth() { return width;}
int Image::getHeight() {return height;}
  
int Image::pixel(int i, int j){
  if(i>=height || j>=width || j<0 || i<0){
    std::cout << width << " " << j << "  +  " << height <<" " << i << std::endl;
    printf("taille incompatible pour acceder a ce pixel\n");
  }else{
    return (image)[i*width+j];
  }
}
	
void Image::setpixel(int i, int j, int couleur){
  if(i>height || j>width || j<0 || i<0){
    printf("taille incompatible pour ecrire dans ce pixel\n");
  }else{
    (image)[i*width+j]=couleur;
  }
}

void Image::save_image(char* fichier){
  TIFF* destFile = TIFFOpen (fichier, "w");
  SavePixels (destFile, image, width, height);
  TIFFClose (destFile);
}

  
Image* Image::copie(int nb){
  Image* res=new Image(nb*width,nb*height);
  int i,j,k,l;
  for(i=0;i<height;i++){
    for(j=0;j<width;j++){
      for(k=0;k<nb;k++){
	for(l=0;l<nb;l++){
	  res->setpixel(i+k*height,j+l*width,this->pixel(i,j));
	}
      }
    }
  }
  
  return res;
}



Image* Image::E2S(){
  int i;
  int j;
  Image* res= new Image(width,height);
  for(i=0;i<width;i++){
    for(j=0;j<height;j++){
      res->setpixel(i,j,i*width+j);
    }
  }
  return res;
}


Image* Image::S2E(Image* exemplaire){
  int i;
  int j;
  Image* res= new Image(width,height);
  for(i=0;i<width;i++){
    for(j=0;j<height;j++){
      int y=this->pixel(i,j)%exemplaire->width;
      int x=(this->pixel(i,j)-y)/exemplaire->width;
      res->setpixel(i,j,exemplaire->pixel(x,y));
    }
  }
  return res;
}

Image* Image::Upsample(){
   
  Image* temp=new Image(2*width, 2*height);
  int i;
  int j;
  for(i=0;i<width;i++){
    for(j=0;j<height;j++){
      temp->setpixel(2*i,2*j,this->pixel(i,j));
      temp->setpixel(2*i+1,2*j,this->pixel(i,j));
      temp->setpixel(2*i,2*j+1,this->pixel(i,j));
      temp->setpixel(2*i+1,2*j+1,this->pixel(i,j));
    }
  }
  return temp;
}

Image* Image::SUpsample(int widthE){
   
  Image* temp=new Image(2*width, 2*height);
  int i;
  int j;
  for(i=0;i<width;i++){
    for(j=0;j<height;j++){
      int coul = 4*(this->pixel(i,j))-2*(this->pixel(i,j)%(widthE/2));
      temp->setpixel(2*i,2*j,(coul+(widthE*widthE))%(widthE*widthE));
      temp->setpixel(2*i+1,2*j,(coul+1+(widthE*widthE))%(widthE*widthE));
      temp->setpixel(2*i,2*j+1,(coul+widthE+(widthE*widthE))%(widthE*widthE));
      temp->setpixel(2*i+1,2*j+1,(coul+widthE+1+(widthE*widthE))%(widthE*widthE));
    }
  }
  return temp;
}

Image* Image::Ssubsample(int widthE){
   
  Image* temp=new Image(width/2, height/2);
  int i;
  int j;
  int x;
  int y;
  int coul;
  for(i=0;i<(width/2);i++){
    for(j=0;j<(height/2);j++){
      coul =this->pixel(2*i,2*j);
      x=coul/(2*widthE);
      y=coul%(2*widthE);
      coul=(x/2)*widthE+(y/2);
      //int coul =((this->pixel(2*i,2*j))/4+((this->pixel(2*i,2*j))%(2*widthE))/4);
      temp->setpixel(i,j,coul);//+(widthE*widthE))%(widthE*widthE));
     
    }
  }
  return temp;
  
}



Image* Image::Subsample(){
   
  Image* temp=new Image(width/2, height/2);
  int i,p1,p2,p3,p4,rouge,vert,bleu,coul;
  int j;
  for(i=0;i<temp->width;i++){
    for(j=0;j<temp->height;j++){ 
      p1=this->pixel(2*i,2*j);
	p2=this->pixel(2*i+1,2*j);
	p3=this->pixel(2*i,2*j+1);
	p4=this->pixel(2*i+1,2*j+1);

	rouge = (((p1 >> 16) & 0x000000FF) +
		 ((p2 >> 16) & 0x000000FF) + 
		 ((p3 >> 16) & 0x000000FF) + 
		 ((p4 >> 16) & 0x000000FF))/4;


	vert = (((p1 >> 8) & 0x000000FF) +
		((p2 >> 8) & 0x000000FF) + 
		((p3 >> 8) & 0x000000FF) + 
		((p4 >> 8) & 0x000000FF))/4;

	bleu = ((p1  & 0x000000FF) +
		(p2  & 0x000000FF) + 
		(p3  & 0x000000FF) + 
		(p4  & 0x000000FF))/4;
	
	coul=((rouge << 16) &0x00FF0000 ) +
	  (( vert << 8)& 0x0000FF00 ) +(( bleu)& 0x000000FF  ) +(( 255 << 24) & 0xFF000000 );
	
	
	temp->setpixel(i,j,coul);
    }
  }
  return temp;
}

Image* Image::Subsample(float div){
   
  Image* temp=new Image( (int) width/div, (int)  height/div);
  int i,p1,p2,p3,p4,rouge,vert,bleu,coul;
  int j;
  for(i=0;i<temp->width;i++){
    for(j=0;j<temp->height;j++){
      p1=this->pixel( (int) (div*i), (int) (div*j));
      p2=this->pixel( (int) (div*i+1), (int) (div*j));
      p3=this->pixel( (int) (div*i), (int) (div*j+1));
      p4=this->pixel( (int) (div*i+1), (int) (div*j+1));

      rouge = (((p1 >> 16) & 0x000000FF) +
	       ((p2 >> 16) & 0x000000FF) + 
	       ((p3 >> 16) & 0x000000FF) + 
	       ((p4 >> 16) & 0x000000FF))/4;


      vert = (((p1 >> 8) & 0x000000FF) +
	      ((p2 >> 8) & 0x000000FF) + 
	      ((p3 >> 8) & 0x000000FF) + 
	      ((p4 >> 8) & 0x000000FF))/4;

      bleu = ((p1  & 0x000000FF) +
	      (p2  & 0x000000FF) + 
	      (p3  & 0x000000FF) + 
	      (p4  & 0x000000FF))/4;
	
      coul=((rouge << 16) &0x00FF0000 ) +
	(( vert << 8)& 0x0000FF00 ) +(( bleu)& 0x000000FF  ) +(( 255 << 24) & 0xFF000000 );
	
	
      temp->setpixel(i,j,coul);
    }
  }
  return temp;
}

void Image::Jitter(Image* exemplaire, int coef){
  int i,j,k,l;
  //int coefJitter = (coef/2)*2+1;
  if (!(coef==0)) {
    for(i=0;i<width;i++){
      for(j=0;j<height;j++){
	this->setpixel(i,j,(this->pixel(i,j)+
			    rand()%(2*coef+1)+
			    rand()%(2*coef+1)*exemplaire->width
			    +2*(exemplaire->width*exemplaire->height))
		       %(exemplaire->width*exemplaire->height));
	
      }
    }
  }
}



int Image::Diffvoisinage(int iE, int jE, int iS, int jS, int k, Image* exemplaire){
  int l, m, x,y;
  int res=0;
  for(l=-k/2;l<k/2+1;l++){
    for(m=-k/2;m<k/2+1;m++){

      y=(this->pixel((iS+l+height)%height,(jS+m+width)%width))%exemplaire->width;
      x=((this->pixel((iS+l+height)%height,(jS+m+width)%width))-y)/exemplaire->width;
      res+=DiffPixels(exemplaire->pixel((iE+l+exemplaire->height)%(exemplaire->height),
					(jE+m+exemplaire->width)%(exemplaire->width)),
		      exemplaire->pixel(x,y));
    }
  }
  return res;
}




void Image::Correction(Image* exemplaire, int taillVoisi){
  int mini=0;
  int minj=0;
  int diffmin=INT_MAX;
  int i,j,k,l,temp;
  for(i=0;i<width;i++){
    for(j=0;j<height;j++){
      for(k=0;k<exemplaire->width;k++){
	for(l=0;l<exemplaire->height;l++){
	    
	  temp=Diffvoisinage(k,l,i,j,taillVoisi,exemplaire);
	  if (temp<diffmin){
	    diffmin=temp;
	    mini=k;
	    minj=l;
	      
	  }
	}
      }

      this->setpixel(i,j,mini*exemplaire->width+minj);
      diffmin=INT_MAX;

    }
  }
  //return res;
}

	



Image* Image::rotate(float degre){
  int i,j;
  int isrc, jsrc;
  int d=(int) sqrtf(Pow(width,2)+Pow(height,2))+1;
  Image* res=new Image(d,d);
  for(i=0;i<d;i++){
    for(j=0;j<d;j++){      
      isrc= (int)(( i-d/2 )*cosf(degre)-(j-d/2)*sinf(degre))+height/2 ;
      jsrc= (int)(( i-d/2 )*sinf(degre)+(j-d/2)*cosf(degre))+width/2 ;
      if (isrc<height && isrc>=0 && jsrc<width && jsrc>=0){
	res->setpixel(i,j,this->pixel(isrc,jsrc));
      }else{
	res->setpixel(i,j,-1);
      }
    }
  }

  return res;
}
	
