#include "loadSaveTiff.h"

typedef unsigned char uint8;


int LoadPixels (TIFF* file, int** memory, int* width, int* height)
{
  int     res;
  
  /* Get the width & height of the image */

  res = TIFFGetField (file, TIFFTAG_IMAGEWIDTH, width);
  if (res != 1) {
    printf ("Tag IMAGEWIDTH was not found in tiff file\n");
    return 1;
  }
  res = TIFFGetField (file, TIFFTAG_IMAGELENGTH, height);
  if (res != 1) {
    printf ("Tag IMAGEHEIGHT was not found in tiff file\n");
    return 1;
  }
  /*printf ("TIFF file has an image of size %ix%i.\n",
                                          (int)(*width), (int)(*height));

  Allocate memory for the pixels and load them from file */

  *memory = (int*)_TIFFmalloc ((*width) * (*height) * sizeof(int));

  if (*memory == (int*)NULL) {
    printf ("Could not allocate memory for image pixels.\n");
    return 1;
  }
  res = TIFFReadRGBAImage (file, *width, *height,(uint32*) *memory, 0);
  if (res != 1) {
    printf ("Could not load image pixels into memory\n");
    _TIFFfree ((tdata_t)(*memory));
    *memory = (int*)NULL;
    return 1;
  }

  return 0;
}



int SwapBuffer (int** memoryD, int* memoryS, int width, int height)
{
  int        i;
  int*    swap;
  int*    cursS;
  int*    cursD;
  int     pixel;
  int        x, y;
	
  swap = (int*)_TIFFmalloc (width * height * sizeof(int));
  if (swap == (int*)NULL) {
    printf ("Could not allocate swap buffer\n");
    return 1;
  }

  cursS = memoryS;
  cursD = swap + (height - 1) * width;
	
  /* Il faut tester le format de la machine : little ou big endian. */
  i = 0;
  *((char *)(&i)) = 1;

  if (i == 1)
  {
    /* En little endian, il ne faut pas echanger les canaux avant la sauvegarde
       car en memoire (PAS dans la variable pixel, mais en representation interne),
       les informations sont rangees dans l'ordre RVBX. 
    */
    for (y = 0; y < height; y++) {
      for (x = 0; x < width; x++) {
        *cursD = *cursS;
        cursS++;
        cursD++;
      }
      cursD -= 2 * width;
    }
  }
  else 
  {
    /* En big endian, il faut echanger les canaux avant la sauvegarde car en memoire,
       les informations sont rangees dans l'ordre XBVR. 
    */
    for (y = 0; y < height; y++) {
      for (x = 0; x < width; x++) {
        pixel = *cursS;
        *cursD = ((pixel & 0xFF000000) >> 24) |
                 ((pixel & 0x00FF0000) >>  8) |
                 ((pixel & 0x0000FF00) <<  8) |
                 ((pixel & 0x000000FF) << 24);
        cursS++;
        cursD++;
      }
      cursD -= 2 * width;
    }
  }

  *memoryD = swap;
  return 0;
}




int SavePixels (TIFF* file, int* memory, int width, int height)
{
  int     res;
  int     dataSize;
  
  res =   TIFFSetField (file, TIFFTAG_IMAGEWIDTH,     width);
  if (res == 1)
    res = TIFFSetField (file, TIFFTAG_IMAGELENGTH,    height);
  if (res == 1)
    res = TIFFSetField (file, TIFFTAG_BITSPERSAMPLE,  (uint16)8);
  if (res == 1)
    res = TIFFSetField (file, TIFFTAG_PHOTOMETRIC,    (uint16)2);
  if (res == 1)
    res = TIFFSetField (file, TIFFTAG_SAMPLESPERPIXEL,(uint16)4);
  if (res == 1)
    res = TIFFSetField (file, TIFFTAG_PLANARCONFIG,
                                              (uint16)PLANARCONFIG_CONTIG);
	if (res == 1) {
    dataSize = sizeof(int) * width * height;
    res = (int)TIFFWriteRawStrip (file, 0, memory, dataSize);
    if (res == -1)
      res = 0;
  }
  if (res == 0) {
    printf ("An error occured while trying to save TIFF file\n");
    return 1;
  }

  return 0;
}
