#include <iostream>
#include <string.h>
#include <string>
#include <stdlib.h> 
#include <ctime>
#include <sstream>
#include <sys/stat.h>
#include <sys/types.h>
#include "fctauxi.h"
#include "image.h"
#include "annUse.h"
#include "vecteur.h"
#include <time.h> 


void help(){
  printf("=====================================================\n");
  printf("Utilisation du programme :\n");
  printf("./synthTextAdvect <option> <fichier exemplaire>\n");
  printf("\n");
  printf("Options : \n");
  printf("-j <int> puissance du Jitter\n");
  printf("-u <int> nombre de sous-passe de upsample\n");
  printf("-n <int> nombre de Corrections pour la génération de la texture\n");
  printf("-y <int> nombre de corrections lors du calcul de l'animation\n");
  printf("-k <int> taille du voisinage de recherche\n");
  printf("-r <int> Coefficient d'aggrandisement de l'exemplaire de depart.\n");
  printf("L'image synthetisé sera de la taille de l'image exemplaire multiplié par 2^r\n");
  printf("-a <int> pour donner de l'entropie a generateur aléatoire\n");
  printf("-m <nom_du_modele> champ de vecteur a appliquer lors de la synthese de texture.\n");
  printf("-d <nom_du_deplacement> <nb images> deplacement constant a appliquer a la texture.\n");
  printf("-t <nom_du_deplacement_temporel> <nb images> deplacement variable a appliquer a la texture\n");
  printf("-s on fait un subsample avant la passe de correction lors du calcul de l'animation\n");
  printf("\n");
  printf("REMARQUE : l'image exemplaire doit être une image carré\n");
  printf("(de fait la largeur et la hauteur de l'image sont égales)\n");
  printf("et de préférence raccordable aux bords.\n");
  printf("=====================================================\n");

}


using namespace std;

//Correction d'une image par rapport a l'exemplaire Exemplaire
void Correction (Image* A, int taillVoisi, annUse* annEx, Image* Exemplaire, Vecteur* Advection){

  int i,j;
  Point* queryPt=new Point(0,0);
  Point* res;
  float angle=0;
  float norme=0;
  for(i=0;i<A->height;i++){
    queryPt->abs=i;
    for(j=0;j<A->width;j++){
      queryPt->ord=j;
      angle=Advection->angle(i,j);
      norme=Advection->norme(i,j);
      res=annEx->bestApproximateMatch(A,queryPt,Exemplaire,angle,norme);
      A->setpixel(i,j,(res->abs)*(Exemplaire->width)+res->ord);
    }
  }
  delete res;
  delete queryPt;
}

//Deplacement de l'image A avec le champ de vecteurs Deplacement
void Deplace(Image* A, Image* Exemplaire, Vecteur* Deplacement){

  int i,j;
  float angle=0;
  float norme=0;
  float fr,lr;
  int w=A->width;
  int wE=Exemplaire->width;
  Image* ret = new Image(w,w);
  for(i=0;i<A->height;i++){
    for(j=0;j<A->width;j++){
      angle=Deplacement->angle(i,j);
      norme=Deplacement->norme(i,j);
      fr= norme*(cosf(angle));
      lr= norme*(sinf(angle));
      if ( round(i-lr)>=w || round(i-lr)<0 ||(round(j+fr)>=w) || round(j+ fr)<0) {
	ret->setpixel(i,j,rand()%(wE*wE));
      }else{ 
	ret->setpixel(i,j,A->pixel((int) round(i-lr),(int) round(j+fr)));
      }
    }
  }
  *A=*ret;

}

//Construction d'un champ de vecteur temporel
Vecteur** ChampsTemporels (string modele, int nbImag, int tailMat){
  
  Vecteur**MatDeplace=(Vecteur**) malloc(nbImag*sizeof(Vecteur*));
  int i;
 

  if (modele=="S"){
    for(i=0;i<nbImag/2;i++){
      MatDeplace[i]=new Vecteur(tailMat,tailMat,"S",i*1000*2/nbImag);
    }
    for(i=nbImag/2;i<nbImag;i++){
      MatDeplace[i]=new Vecteur(tailMat,tailMat,"S",1000);
    }
  }else if (modele=="hypnose") {
    for(i=0;i<nbImag/4;i++){
      MatDeplace[i]=new Vecteur(tailMat,tailMat,"puit-1",0);
    }
    for(i=nbImag/4;i<nbImag/2;i++){
      MatDeplace[i]=new Vecteur(tailMat,tailMat,"puit1",0);
    }
    for(i=nbImag/2;i<3*nbImag/4;i++){
      MatDeplace[i]=new Vecteur(tailMat,tailMat,"puit-1",0);
    }
    for(i=3*nbImag/4;i<nbImag;i++){
      MatDeplace[i]=new Vecteur(tailMat,tailMat,"puit1",0);
    }
  }else if (modele=="hypnose_tournant") {
    for(i=0;i<nbImag/7;i++){
      MatDeplace[i]=new Vecteur(tailMat,tailMat,"puit-1",0);
    }
    for(i=nbImag/7;i<2*nbImag/7;i++){
      MatDeplace[i]=new Vecteur(tailMat,tailMat,"cercle",0);
    }
    for(i=2*nbImag/7;i<3*nbImag/7;i++){
      MatDeplace[i]=new Vecteur(tailMat,tailMat,"puit1",0);
    }
    for(i=3*nbImag/7;i<4*nbImag/7;i++){
      MatDeplace[i]=new Vecteur(tailMat,tailMat,"cercle",0);
    }  
    for(i=4*nbImag/7;i<5*nbImag/7;i++){
      MatDeplace[i]=new Vecteur(tailMat,tailMat,"puit-1",0);
    }  
    for(i=5*nbImag/7;i<6*nbImag/7;i++){
      MatDeplace[i]=new Vecteur(tailMat,tailMat,"cercle",0);
    }  
    for(i=6*nbImag/7;i<nbImag;i++){
      MatDeplace[i]=new Vecteur(tailMat,tailMat,"puit1",0);
    }
  }else if (modele=="null") {
    MatDeplace[0]=new Vecteur(tailMat,tailMat,"null",0);
    for(i=1;i<nbImag;i++){
      MatDeplace[i]= MatDeplace[0];
    }
  }else{
    MatDeplace[0]=new Vecteur(tailMat,tailMat,"null",0);
    for(i=1;i<nbImag;i++){
      MatDeplace[i]= MatDeplace[0];
    }
  }
  return MatDeplace;
 
}




int main (int argc, char* argv[])
{  

  /*Options : 
    -j <int> puissance du Jitter
    -u <int> nombre de sous-passe de upsample
    -n <int> nombre de Corrections pour la génération de la texture
    -y <int> nombre de corrections lors du calcul de l'animation
    -k <int> taille du voisinage de recherche
    -r <int> Coefficient d'aggrandisement de l'exemplaire de depart.
    L'image synthetisé sera de la taille de l'image exemplaire multiplié par 2^r
    -a <int> pour donner de l'entropie a generateur aléatoire
    -m <nom_du_modele> champ de vecteur a appliquer lors de la synthese de texture.
    -d <nom_du_deplacement> <nb images> deplacement constant a appliquer a la texture.
    -t <nom_du_deplacement_temporel> <nb images> deplacement variable a appliquer a la texture
    -s on fait un subsample avant la passe de correction lors du calcul de l'animation
    
    mettre en dernier le nom du fichier exemplaire
  */
  if (argc==1){
    help();
    return 1;
  }
  
  long Htoutdebut;
  time(&Htoutdebut);


  printf("\n                            ================================= \n");
  printf("\n                            | Synthetiseur Amat Barjon !!!! |\n");
  printf("\n                            ================================= \n");
  
  int coefJitter = 3;
  int nbUpsamp = 3;
  int nbCorrecti = 2;
  int nbCorrectiMouv = 1;
  int taillVoisi = 5;
  int coefAgrandi = 2;
  int srandd = 100;
  int nbdeplace=0;
  int nbdepTemp=0;
  string modele="null";
  string deplace="null";
  string depTemp="null";
  int t=0;
  int subsampl=0;
  int i;
  int j;


  //gestion des options
  for (i = 1; i < argc; i++) {
    if (!strcmp(argv[i], "-help")){
      help();
      return 1;
    }
    if (!strcmp(argv[i], "-j")) coefJitter=atoi(argv[i+1]);
    if (!strcmp(argv[i], "-u")) nbUpsamp=atoi(argv[i+1]);
    if (!strcmp(argv[i], "-n")) nbCorrecti=atoi(argv[i+1]);
    if (!strcmp(argv[i], "-y")) nbCorrectiMouv=atoi(argv[i+1]);
    if (!strcmp(argv[i], "-k")) taillVoisi=atoi(argv[i+1]);
    if (!strcmp(argv[i], "-r")) coefAgrandi=atoi(argv[i+1]);
    if (!strcmp(argv[i], "-a")) srandd=atoi(argv[i+1]);
    if (!strcmp(argv[i], "-s")) subsampl=1;
    if (!strcmp(argv[i], "-m")){
      modele=argv[i+1];    
    }
    if (!strcmp(argv[i], "-d")){
      deplace=argv[i+1];
      nbdeplace=atoi(argv[i+2]);
    }
    if (!strcmp(argv[i], "-t")){
      depTemp=argv[i+1];
      t=1;
      nbdepTemp=atoi(argv[i+2]);
      nbdeplace=nbdepTemp;
    }
    
  }

  //initialisation du rand
  if (srandd==100){
    srand(coefJitter*nbUpsamp*nbCorrecti*taillVoisi*coefAgrandi+12);
  }else{
    srand(srandd);
  }

  //initialisation des exemplaires et de ANN
  Image** Exemplaires=(Image**) malloc(nbUpsamp*sizeof(Image*));
  Exemplaires[0]=new Image(argv[argc-1]);

  annUse** annE=(annUse**) malloc(nbUpsamp*sizeof(annUse*));
  annE[0]= new annUse(Exemplaires[0], taillVoisi);
 
  //Matrices de vecteurs pour l'advection(une matrice par niveau de la pyramide)

  Vecteur**MatVects=(Vecteur**) malloc(nbUpsamp*sizeof(Vecteur*));
  int tailMat=Pow(2,coefAgrandi)*(Exemplaires[0]->width);
  MatVects[0]=new Vecteur(tailMat,tailMat,modele);
  
  for(i=1;i<=nbUpsamp;i++){
    MatVects[i]=MatVects[i-1]->Subsample(); 
  }

  //calcule des exemplaires et des espaces ANN(un par niveau de la pyramide)

  for(i=1;i<=nbUpsamp;i++){
    Exemplaires[i]=Exemplaires[i-1]->Subsample(); 
    annE[i]= new annUse(Exemplaires[i], taillVoisi);
  
  }
  

 
  //calcul de la taille de S-1
  int tailL, tailH;
  tailL=(Exemplaires[0]->width)/Pow(2,nbUpsamp)*Pow(2,coefAgrandi);
  tailH=(Exemplaires[0]->height)/Pow(2,nbUpsamp)*Pow(2,coefAgrandi);
  
  //creation de S-1
  Image* SMoinsUn=new Image(tailL,tailH);
  for(i=0;i<SMoinsUn->width;i++){
    for(j=0;j<SMoinsUn->height;j++){
      SMoinsUn->setpixel(i,j,0);
    }
  }

 
  Image* S=SMoinsUn;
  
  //gestion des noms des images et des noms des dossiers qui seront créés lors de la génération
  string nomImage=argv[argc-1];
  char* pwdImage=argv[argc-1];
  int longeur=nomImage.length();
  nomImage="";
  for (i=longeur-5;i>0; i--){
    if ('/'==pwdImage[i]) break;
    nomImage=pwdImage[i]+nomImage;
  }

  string nomOptions="";
  for(i=1;i<argc-1;i++){
    nomOptions+=argv[i];
  }
  
  string nomDossier =nomImage+nomOptions;
  mkdir(nomDossier.c_str(),01777);

  string nomSynth;
  string extension=".tif";
  string numU,numN;
  numU="0";
  int k;

  long H,Hdebut;

  //DEBUT DE LA GENERATION DE LA TEXTURE
  printf("\n                  GENERATION DE LA TEXTURE\n\n");
  for(i=0;i<nbUpsamp;i++){
    std::cout << "Upsample numero " << i+1 <<" :" << std::endl;
    numU=itoa(i);
    Exemplaires[nbUpsamp-i-1]->save_image((char*) (nomDossier+"/"+"Exemplaire"+nomImage+"u"+numU+extension).c_str());
    //upsampling et Jitter
    S=S->SUpsample(Exemplaires[nbUpsamp-i-1]->width);
    S->Jitter(Exemplaires[nbUpsamp-i-1],coefJitter/(Pow((i+1),3)));
    S->S2E(Exemplaires[nbUpsamp-i-1])->save_image((char*) (nomDossier+"/"+nomImage+"u"+numU+extension).c_str());
    //passes de corrections
    for(j=0;j<nbCorrecti;j++){
      time(&Hdebut);
      printf("     Correction numero %d...",j+1);
      fflush(stdout);
      numN=itoa(j+1);
      Correction(S,taillVoisi,annE[nbUpsamp-i-1],Exemplaires[nbUpsamp-i-1], MatVects[nbUpsamp-i-1]);
      time(&H);
      std::cout<<"effectué en : "<< (H-Hdebut)/60 <<"m"<<(H-Hdebut)%60<<"sec, ";
      std::cout<<"terminé a : "<< localtime(&H)->tm_hour<<":"<< localtime(&H)->tm_min<<":"<< localtime(&H)->tm_sec<< "\n";
      S->S2E(Exemplaires[nbUpsamp-i-1])->save_image((char*) (nomDossier+"/"+nomImage+"u"+numU+"n"+numN+extension).c_str());
    }
  }
  //sauvegarde du resultat
  Image* Fin=S;
  Image* testi=new Image(S->width,S->width);
  Fin=S->S2E(Exemplaires[0]);
  Fin->save_image((char*) (nomImage+nomOptions+extension).c_str());


  if(nbdeplace==0){
    //si on ne genere pas d'animation, on s'arrete
    system(("display "+nomImage+nomOptions+extension+" &").c_str());
    return 1;//si on veut pas faire d'animations on s'arrete
  }

  //creation du dossier anim pour l'animation
  mkdir((nomDossier+"/anim").c_str(),01777);
  //fichier de depart
  Fin->save_image((char*) (nomDossier+"/anim/000"+extension).c_str());
  //Matrice de déplacement !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  int nbIter;
  (t==1) ? nbIter=nbdepTemp : nbIter=nbdeplace;
  
  Vecteur**MatDeplace;
  


  //Matrices de vecteurs pour l'animation(une matrice par niveau de la pyramide)
  if (t==1){ 
    MatDeplace=ChampsTemporels(depTemp,nbdepTemp,tailMat);
  }else{
    MatDeplace=(Vecteur**) malloc(nbIter*sizeof(Vecteur*));
    MatDeplace[0]=new Vecteur(tailMat,tailMat,deplace,0);
    for (i=1;i<nbIter;i++){
      MatDeplace[i]= MatDeplace[0];
    }
  }
  printf("\n                  GENERATION DE L'ANIMATION\n\n");
  std::cout<<"Temps estimé : "<< ((H-Hdebut)*(nbCorrectiMouv*nbIter))/(60*60) <<"h"
	   << (((H-Hdebut)*(nbCorrectiMouv*nbIter))%3600)/60  <<"m"
	   <<((H-Hdebut)*(nbCorrectiMouv*nbIter))%60<<"sec\n";
 
  for (k=0;k<nbIter;k++){
    time(&Hdebut);
    printf("Calcul de l'image numero %d...",k+2);
    fflush(stdout);
    numU=itoa(k+1);
    if (k+1<10){
      numU="00"+numU;
    }else if (k+1<100){
      numU="0"+numU;
    }
    Deplace(S,Exemplaires[0],MatDeplace[k]);
    
    if (subsampl==1) {
      S=S->Ssubsample(Exemplaires[1]->width);
      
      
      for(j=0;j<nbCorrectiMouv;j++){
	Correction(S,taillVoisi,annE[1],Exemplaires[1], MatVects[1]);
      }
      S=S->SUpsample(Exemplaires[0]->width);
    }
    
    for(j=0;j<nbCorrectiMouv;j++){
      Correction(S,taillVoisi,annE[0],Exemplaires[0], MatVects[0]);
    }
    time(&H);
    std::cout<<"effectué en : "<< (H-Hdebut)/60 <<"m"<<(H-Hdebut)%60<<"sec, ";
    std::cout<<"terminé a : "<< localtime(&H)->tm_hour<<":"<< localtime(&H)->tm_min<<":"<< localtime(&H)->tm_sec<< "\n";

    Fin=S->S2E(Exemplaires[0]);
    Fin->save_image((char*) (nomDossier+"/anim/"+numU+extension).c_str());
    
  }
  time(&H);
  std::cout<<"Programme effectué en : "<< (H-Htoutdebut)/(60*60) <<"h"<<((H-Htoutdebut)%3600)/60 <<"m"<<(H-Htoutdebut)%60<<"sec \n"<<endl;
  
  system(("animate -delay 5 "+ nomDossier+"/anim/*.tif &").c_str()) ;
  
  
  return 1;
  
}
